using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.IO;
using System.Diagnostics;

namespace BucketAndSpade {
	class Program {

		#region Tool Paths

		private static string XslTransformPath;
		private static string MRefBuilderPath;
		private static string BuildAssemblerPath;

		#endregion

		static void Main(string[] args) {

			// Display header and instructions (if applicable).
			AssemblyName AppName = Assembly.GetExecutingAssembly().GetName();
			Console.WriteLine(AppName.Name + " " + AppName.Version);
			if (args.Length < 1 || args.Length > 2) {
				Console.WriteLine("Usage: BucketSpade assembly.dll [assembly.xml]");
				return;
			}

			#region Path Setup

			// Check tool paths:
			if (!Directory.Exists(Properties.Settings.Default.SandcastlePath)) {
				Console.WriteLine("Could not find Sandcastle's path.");
				return;
			}
			if (!Directory.Exists(Properties.Settings.Default.HtmlHelpWorkshopPath)) {
				Console.WriteLine("Could not find HTML Help Workshop's path.");
				return;
			}

			MRefBuilderPath = Path.Combine(Properties.Settings.Default.SandcastlePath, @"ProductionTools\MRefBuilder.exe");
			if (!File.Exists(MRefBuilderPath)) {
				Console.WriteLine("Couldn't find MRefBuilder.exe");
				return;
			}

			XslTransformPath = Path.Combine(Properties.Settings.Default.SandcastlePath, @"ProductionTools\XslTransform.exe");
			if (!File.Exists(XslTransformPath)) {
				Console.WriteLine("Couldn't find XslTransform.exe");
				return;
			}

			string CopyOutputPath = Path.Combine(Properties.Settings.Default.SandcastlePath, @"Presentation\vs2005\copyOutput.bat");
			if (!File.Exists(CopyOutputPath)) {
				Console.WriteLine("Couldn't find copyOutput.bat");
				return;
			}

			BuildAssemblerPath = Path.Combine(Properties.Settings.Default.SandcastlePath, @"ProductionTools\BuildAssembler.exe");
			if (!File.Exists(BuildAssemblerPath)) {
				Console.WriteLine("Couldn't find BuildAssembler.exe");
				return;
			}

			string OutputConfiguration = Path.Combine(Properties.Settings.Default.SandcastlePath, @"Presentation\vs2005\configuration\sandcastle.config");
			if (!File.Exists(OutputConfiguration)) {
				Console.WriteLine("Couldn't find sandcastle.config");
				return;
			}

			string HtmlHelpCompiler = Path.Combine(Properties.Settings.Default.HtmlHelpWorkshopPath, @"hhc.exe");
			if (!File.Exists(HtmlHelpCompiler)) {
				Console.WriteLine("Couldn't find hhc.exe");
				return;
			}

			// Check parameter paths:
			string AssemblyName = Path.GetFullPath(args[0]);
			string DocumentationName = Path.GetFullPath(args.Length == 2 ? args[1] : Path.Combine(Path.GetDirectoryName(args[0]), Path.GetFileNameWithoutExtension(args[0]) + ".xml"));

			if (!File.Exists(AssemblyName)) {
				Console.WriteLine("Could not find assembly '" + AssemblyName + "'.");
				return;
			}
			if (!File.Exists(DocumentationName)) {
				Console.WriteLine("Could not find assembly's XML documentation '" + DocumentationName + "'.");
				return;
			}

			bool CreatedCommentsXml = false;

			if (Path.GetFileName(DocumentationName).ToLowerInvariant() != "comments.xml") {
				string SourceDocumentationName = DocumentationName;
				DocumentationName = Path.Combine(Path.GetDirectoryName(DocumentationName), "comments.xml");
				try {
					if (File.Exists(DocumentationName)) File.Delete(DocumentationName);
					File.Copy(SourceDocumentationName, DocumentationName);
					CreatedCommentsXml = true;
				} catch (Exception ex) {
					Console.WriteLine("Couldn't create comments.xml: " + ex.Message);
					return;
				}
			}

			#endregion

			#region Output Setup

			// Let's go crazy.

			// Start by creating the output directory:
			string OutputDirectory = Path.GetDirectoryName(AssemblyName);
			if (!Directory.Exists(OutputDirectory)) {
				Console.WriteLine("Error finding output directory.");
				return;
			}

			Console.WriteLine("Creating Output Directory Structure...");
			Directory.SetCurrentDirectory(OutputDirectory);
			ExecuteCommandLineTool(CopyOutputPath, "", false);

			#endregion


			#region Compilation

			Console.WriteLine("Performing Reflection...");
			string ReflectionOrg = Path.Combine(OutputDirectory, "reflection.org");
			if (ExecuteCommandLineTool(MRefBuilderPath, string.Format("\"{0}\" /out:\"{1}\"", AssemblyName, ReflectionOrg)) != 0) return;


			Directory.SetCurrentDirectory(OutputDirectory);

			Console.WriteLine("Transforming Reflection...");
			string ReflectionXml = Path.Combine(OutputDirectory, "reflection.xml");
			
			if (!XslTransform(ReflectionOrg, ReflectionXml, 
				Path.Combine(Properties.Settings.Default.SandcastlePath, @"ProductionTransforms\ApplyVSDocModel.xsl"),
				Path.Combine(Properties.Settings.Default.SandcastlePath, @"ProductionTransforms\AddFriendlyFilenames.xsl")
			)) return;

			Console.WriteLine("Creating Topic Manifest...");
			string ManifestXml = Path.Combine(OutputDirectory, "manifest.xml");
			if (!XslTransform(ReflectionXml, ManifestXml, Path.Combine(Properties.Settings.Default.SandcastlePath, @"ProductionTransforms\ReflectionToManifest.xsl"))) return;

			Console.WriteLine("Generating HTML Output...");
			if (ExecuteCommandLineTool(BuildAssemblerPath, string.Format(@"/config:""{0}"" ""{1}""", OutputConfiguration, ManifestXml), false) != 0) return;

			Console.WriteLine("Creating HTML Help Project...");
			string HtmlHelpHhp = Path.Combine(OutputDirectory, Path.Combine("Output", "test.hhp"));
			if (!XslTransform(ReflectionXml, HtmlHelpHhp, Path.Combine(Properties.Settings.Default.SandcastlePath, @"ProductionTransforms\ReflectionToChmProject.xsl"))) return;

			Console.WriteLine("Creating Table of Contents...");
			string TocXml = Path.Combine(OutputDirectory, "toc.xml");
			if (!XslTransform(ReflectionXml, TocXml, Path.Combine(Properties.Settings.Default.SandcastlePath, @"ProductionTransforms\createvstoc.xsl"))) return;		
			

			Console.WriteLine("Generating HTML Help Project Contents...");
			string HtmlHelpHhc = Path.Combine(OutputDirectory, Path.Combine("Output", "test.hhc"));
			if (!XslTransform(TocXml, HtmlHelpHhc, Path.Combine(Properties.Settings.Default.SandcastlePath, @"ProductionTransforms\TocToChmContents.xsl"))) return;

			Console.WriteLine("Generating HTML Help Project Index...");
			string HtmlHelpHhk = Path.Combine(OutputDirectory, Path.Combine("Output", "test.hhk"));
			if (!XslTransform(ReflectionXml, HtmlHelpHhk, Path.Combine(Properties.Settings.Default.SandcastlePath, @"ProductionTransforms\ReflectionToChmIndex.xsl"))) return;

			Console.WriteLine("Compiling HTML Help...");
			ExecuteCommandLineTool(HtmlHelpCompiler, string.Format(@"""{0}""", HtmlHelpHhp));

			#endregion

			#region Clean Up

			SafeDeleteFile(HtmlHelpHhp);
			SafeDeleteFile(HtmlHelpHhc);
			SafeDeleteFile(HtmlHelpHhk);
			SafeDeleteFile(ReflectionOrg);
			SafeDeleteFile(ReflectionXml);
			SafeDeleteFile(TocXml);
			SafeDeleteFile(ManifestXml);
			SafeMove(Path.Combine(OutputDirectory, Path.Combine("Output", "test.chm")), Path.Combine(OutputDirectory, Path.GetFileNameWithoutExtension(AssemblyName).Replace("#", "Sharp") + ".chm"));
			SafeDeleteDirectory(Path.Combine(OutputDirectory, "Output"));
			SafeDeleteDirectory(Path.Combine(OutputDirectory, "Intellisense"));
			if (CreatedCommentsXml) SafeDeleteFile(DocumentationName);

			#endregion

			Console.WriteLine("Done!");

		}

		#region "Safe" File Operations

		private static void SafeDeleteFile(string filename) {
			try {
				File.Delete(filename);
			} catch (Exception ex) {
				Console.WriteLine("Couldn't delete file: " + ex.Message);
			}
		}

		private static void SafeDeleteDirectory(string path) {
			try {
				Directory.Delete(path, true);
			} catch (Exception ex) {
				Console.WriteLine("Couldn't delete directory: " + ex.Message);
			}
		}

		private static void SafeMove(string source, string destination) {
			try {
				File.Move(source, destination);
			} catch (Exception ex) {
				Console.WriteLine("Couldn't move file: " + ex.Message);
			}
		}

		#endregion

		#region Command-Line Tool Helpers

		private static bool XslTransform(string input, string output, string xslFile) {
			return ExecuteCommandLineTool(XslTransformPath, string.Format(@"/xsl:""{0}"" ""{1}"" /out:""{2}""", xslFile, input, output), false) == 0;
		}

		private static bool XslTransform(string input, string output, string xslFile1, string xslFile2) {
			//HACK: arg:
			return ExecuteCommandLineTool(XslTransformPath, string.Format(@"/xsl:""{0}"" ""{2}"" /xsl:""{1}"" /out:""{3}"" /arg:IncludeAllMembersTopic=true", xslFile1, xslFile2, input, output), false) == 0;
		}

		#endregion

		#region Command-Line Tool Execution

		private static int ExecuteCommandLineTool(string path, string arguments) {
			return ExecuteCommandLineTool(path, arguments, true);
		}

		private static int ExecuteCommandLineTool(string path, string arguments, bool setWorkingDirectory) {
			bool Verbose = Properties.Settings.Default.Verbose;
			using (Process P = new Process()) {

				// Set some environment variables:
				P.StartInfo.FileName = path;
				if (setWorkingDirectory) P.StartInfo.WorkingDirectory = Path.GetDirectoryName(path);
				P.StartInfo.Arguments = arguments;
				P.StartInfo.UseShellExecute = false;
				P.StartInfo.RedirectStandardOutput = true;
				P.StartInfo.RedirectStandardError = true;
				string Prefix;
				DataReceivedEventHandler D = delegate(object sender, DataReceivedEventArgs e) {
					Prefix = ((string)(e.Data ?? "")).ToLowerInvariant();
					if (Verbose || (e.Data != null && (Prefix.StartsWith("warn:") || Prefix.StartsWith("error:")))) {
						Console.WriteLine(e.Data);
					}
				};
				P.OutputDataReceived += new DataReceivedEventHandler(D);
				P.ErrorDataReceived += new DataReceivedEventHandler(D);
				P.Start();
				P.BeginOutputReadLine();
				P.BeginErrorReadLine();
				P.WaitForExit();
				return P.ExitCode;
			}
		}
		#endregion
	}
}
